using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using Utils.ImageProcessing.Exceptions;

namespace ImageProcessing
{
    /// <summary>
    /// Convolution Matrix
    /// </summary>
    internal class ConvMatrix
    {
        public int TopLeft = 0, TopMid = 0, TopRight = 0;
        public int MidLeft = 0, Pixel = 1, MidRight = 0;
        public int BottomLeft = 0, BottomMid = 0, BottomRight = 0;
        public int Factor = 1;
        public int Offset = 0;
        public void SetAll(int nVal)
        {
            TopLeft = TopMid = TopRight = MidLeft = Pixel = MidRight = BottomLeft = BottomMid = BottomRight = nVal;
        }
    }

    /// <summary>
    /// Image processing class
    /// </summary>
    public class ImageProcessor
    {
        #region Variables
        /// <summary>
        /// Image source of the image processing
        /// </summary>
        private Bitmap vBitmap;

        /// <summary>
        /// Pixel's data
        /// </summary>
        private BitmapData vBitmapData;

        /// <summary>
        /// Indicates if we can process the image
        /// </summary>
        private bool vProcessable = true;

        /// <summary>
        /// Pixel format of the image
        /// </summary>
        private PixelFormat vPixelFormat = PixelFormat.Undefined;
        #endregion

        #region Constructor
        /// <summary>
        /// Constructor
        /// </summary>
        public ImageProcessor() { }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="_Bitmap">Image source withon we process</param>
        public ImageProcessor(Bitmap _Bitmap)
        {
            vBitmap = (Bitmap)_Bitmap.Clone();
            vPixelFormat = _Bitmap.PixelFormat;
        }
        #endregion

        #region Methods
        /// <summary>
        /// Locks the bits of the image and prepare the image for processing
        /// </summary>
        public void Start()
        {
            Start(vPixelFormat);
        }
        /// <summary>
        /// Locks the bits of the image and prepare the image for processing with the chosen PixelFormat
        /// </summary>
        /// <param name="_PixelFormat">PixelFormat of the image</param>
        public void Start(PixelFormat _PixelFormat)
        {
            if (!this.vProcessable)
                throw new LockedBitmapException();
            vProcessable = false;

            vBitmapData = vBitmap.LockBits(new Rectangle(0, 0, vBitmap.Width, vBitmap.Height),
                ImageLockMode.ReadWrite, _PixelFormat);
        }
        /// <summary>
        /// Unlocks the bits of the image and close the processing of the image
        /// </summary>
        public void End()
        {
            End(vBitmapData.PixelFormat);
        }
        /// <summary>
        /// Unlocks the bits of the image and close the processing of the image
        /// </summary>
        /// <param name="_PixelFormat">Pixel format of the image</param>
        public void End(PixelFormat _PixelFormat)
        {
            if (!this.vProcessable)
            {
                vBitmapData.PixelFormat = _PixelFormat;
                vBitmap.UnlockBits(vBitmapData);
                vProcessable = true;
            }
        }
        #region Processing
        /// <summary>
        /// Transform the color image into grayscale image
        /// </summary>
        public unsafe void GrayScale()
        {
            IntPtr Scan0 = vBitmapData.Scan0;
            byte* p = (byte*)(void*)Scan0;
            int nOffset = vBitmapData.Stride - vBitmapData.Width * 3;

            //Color values 
            byte Red, Green, Blue;

            for (int y = 0; y < vBitmapData.Height; y++)
            {
                for (int x = 0; x < vBitmapData.Width; x++)
                {
                    Blue = p[0];
                    Green = p[1];
                    Red = p[2];
                    p[0] = p[1] = p[2] = (byte)(.299 * Red + .587 * Green + .114 * Blue);

                    p += 3;
                }
                p += nOffset;
            }
        }
        /// <summary>
        /// Process a median fiter on the image
        /// </summary>
        [Obsolete("Does not process the 24bits images correctly")]
        public unsafe void MedianFilter()
        {
            IntPtr Scan0 = vBitmapData.Scan0;
            byte* p = (byte*)(void*)Scan0;
            int nOffset = vBitmapData.Stride - vBitmapData.Width * 3;
            int[] Matrix = new int[9];

            p -= nOffset - 3;
            for (int y = 0; y < vBitmapData.Height; y++)
            {
                for (int x = 0; x < vBitmapData.Width; x++)
                {
                    //Pixel processing of the matrix. Filling of the matrix

                    //x   y   z
                    Matrix[0] = *(p - (nOffset - 3)); //x                    
                    Matrix[1] = *(p - nOffset); //y
                    Matrix[2] = *(p - (nOffset + 3)); //z

                    // x   o   y
                    Matrix[3] = *(p - 3); //x
                    Matrix[4] = *p;
                    Matrix[5] = *(p + 3); //z

                    // x'  y'  z'
                    Matrix[6] = *(p + (nOffset - 3)); //x'
                    Matrix[7] = *(p + nOffset); //y'
                    Matrix[8] = *(p + (nOffset + 3)); //z'

                    //Sorting of the pixels
                    int buffer;
                    for (int i = 8; i >= 0; i--)
                    {
                        for (int j = 0; j < i; j++)
                        {
                            if (Matrix[i] < Matrix[j + 1])
                            {
                                buffer = Matrix[j];
                                Matrix[j] = Matrix[j + 1];
                                Matrix[j + 1] = buffer;
                            }
                        }
                    }
                    p[0] = (byte)Matrix[4];
                    p += 3;

                }
                p += nOffset;
            }
        }
        /// <summary>
        /// Execute a Sobel fitler for edge detection
        /// </summary>
        public void SobelFilter()
        {
            ConvMatrix m = new ConvMatrix();
            m.SetAll(0);
            m.TopLeft = m.BottomLeft = 1;
            m.TopRight = m.BottomRight = -1;
            m.MidLeft = 2;
            m.MidRight = -2;
            m.Offset = 0;
            Convolution(m);
        }
        /// <summary>
        /// Execute a Prewitt fitler for edge detection
        /// </summary>
        public void PrewittFilter()
        {
            ConvMatrix m = new ConvMatrix();
            m.SetAll(0);
            m.TopLeft = m.MidLeft = m.BottomLeft = -1;
            m.TopRight = m.MidRight = m.BottomRight = 1;
            m.Offset = 0;
            Convolution(m);
        }
        /// <summary>
        /// See the difference with another picture
        /// </summary>
        /// <param name="_ToCompare">The image to compare to the other</param>
        public unsafe int Difference(Bitmap _ToCompare)
        {
            return Difference(_ToCompare, vPixelFormat);
        }
        /// <summary>
        /// See the difference with another picture
        /// </summary>
        /// <param name="_ToCompare">The image to compare to the other</param>
        /// <param name="_PixelFormat">Pixel format of the two images</param>
        public unsafe int Difference(Bitmap _ToCompare, PixelFormat _PixelFormat)
        {
            return Difference(_ToCompare, _PixelFormat, 40);
        }
        /// <summary>
        /// See the difference with another picture
        /// </summary>
        /// <param name="_ToCompare">The image to compare to the other</param>
        /// <param name="_PixelFormat">Pixel format of the two images</param>
        /// <param name="_Sensibility">Color difference within there is no alarm </param>
        public unsafe int Difference(Bitmap _ToCompare, PixelFormat _PixelFormat, int _Sensibility)
        {
            int PixelsDifferent = 0;

            #region ToCompare
            BitmapData vBitmapDataBis = _ToCompare.LockBits(new Rectangle(0, 0, _ToCompare.Width, _ToCompare.Height),
                ImageLockMode.ReadWrite, _PixelFormat);
            IntPtr Scan02 = vBitmapDataBis.Scan0;
            byte* p2 = (byte*)(void*)Scan02;
            #endregion

            IntPtr Scan0 = vBitmapData.Scan0;
            byte* p = (byte*)(void*)Scan0;
            int nOffset = vBitmapData.Stride - vBitmapData.Width * 3;

            if ((vBitmapDataBis.Width * vBitmapDataBis.Height) != (vBitmapData.Width * vBitmapData.Height)
                || vBitmapData.PixelFormat != _PixelFormat)//vBitmapDataBis.PixelFormat)
                throw new FormatBitmapException("The two images to compare are different");

            for (int y = 0; y < vBitmapData.Height; y++)
            {
                for (int x = 0; x < vBitmapData.Width; x++)
                {
                    //Process the difference
                    if (p[0] < p2[0] - _Sensibility || p[0] > p2[0] + _Sensibility)
                    {
                        PixelsDifferent++;
                        p[0] = 0; //Blue
                        p[1] = 0; //Green
                        p[2] = 255; //Red
                    }
                    else
                    {
                        p[0] = 0; //Blue
                        p[1] = 0; //Green
                        p[2] = 0; //Red
                    }
                    p += 3;
                    p2 += 3;
                }
                p += nOffset;
                p2 += nOffset;
            }
            _ToCompare.UnlockBits(vBitmapDataBis);
            return PixelsDifferent;
        }
        /// <summary>
        /// Process a convolution on the image
        /// </summary>
        /// <param name="_Matrix">Convolution Matrix</param>
        private unsafe void Convolution(ConvMatrix _Matrix)
        {
            Bitmap ClonedImage = (Bitmap)vBitmap.Clone();
            BitmapData ClonedBitmapData = ClonedImage.LockBits(new Rectangle(0, 0, ClonedImage.Width, ClonedImage.Height),
                ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            IntPtr Scan0 = vBitmapData.Scan0;
            IntPtr ClonedScan0 = ClonedBitmapData.Scan0;
            int Stride = ClonedBitmapData.Stride;
            int ClonedStride = Stride * 2;

            byte* Ptr = (byte*)(void*)Scan0;
            byte* ClonedPtr = (byte*)(void*)ClonedScan0;

            int nOffset = Stride - vBitmap.Width * 3;
            int nWidth = vBitmap.Width - 2;
            int nHeight = vBitmap.Height - 2;

            int nPixel;

            for (int y = 0; y < nHeight; ++y)
            {
                for (int x = 0; x < nWidth; ++x)
                {
                    nPixel = ((((ClonedPtr[2] * _Matrix.TopLeft) + (ClonedPtr[5] * _Matrix.TopMid) + (ClonedPtr[8]
                        * _Matrix.TopRight) + (ClonedPtr[2 + Stride] * _Matrix.MidLeft) + (ClonedPtr[5 + Stride]
                        * _Matrix.Pixel) + (ClonedPtr[8 + Stride] * _Matrix.MidRight) + (ClonedPtr[2 + ClonedStride]
                        * _Matrix.BottomLeft) + (ClonedPtr[5 + ClonedStride] * _Matrix.BottomMid) + (ClonedPtr[8 + ClonedStride]
                        * _Matrix.BottomRight)) / _Matrix.Factor) + _Matrix.Offset);

                    if (nPixel < 0) nPixel = 0;
                    if (nPixel > 255) nPixel = 255;

                    Ptr[5 + Stride] = (byte)nPixel;

                    nPixel = ((((ClonedPtr[1] * _Matrix.TopLeft) + (ClonedPtr[4] * _Matrix.TopMid) + (ClonedPtr[7]
                        * _Matrix.TopRight) + (ClonedPtr[1 + Stride] * _Matrix.MidLeft) + (ClonedPtr[4 + Stride]
                        * _Matrix.Pixel) + (ClonedPtr[7 + Stride] * _Matrix.MidRight) + (ClonedPtr[1 + ClonedStride]
                        * _Matrix.BottomLeft) + (ClonedPtr[4 + ClonedStride] * _Matrix.BottomMid) + (ClonedPtr[7 + ClonedStride]
                        * _Matrix.BottomRight)) / _Matrix.Factor) + _Matrix.Offset);

                    if (nPixel < 0) nPixel = 0;
                    if (nPixel > 255) nPixel = 255;

                    Ptr[4 + Stride] = (byte)nPixel;

                    nPixel = ((((ClonedPtr[0] * _Matrix.TopLeft) + (ClonedPtr[3] * _Matrix.TopMid) + (ClonedPtr[6]
                        * _Matrix.TopRight) + (ClonedPtr[0 + Stride] * _Matrix.MidLeft) + (ClonedPtr[3 + Stride]
                        * _Matrix.Pixel) + (ClonedPtr[6 + Stride] * _Matrix.MidRight) + (ClonedPtr[0 + ClonedStride]
                        * _Matrix.BottomLeft) + (ClonedPtr[3 + ClonedStride] * _Matrix.BottomMid) + (ClonedPtr[6 + ClonedStride]
                        * _Matrix.BottomRight)) / _Matrix.Factor) + _Matrix.Offset);

                    if (nPixel < 0) nPixel = 0;
                    if (nPixel > 255) nPixel = 255;

                    Ptr[3 + Stride] = (byte)nPixel;

                    Ptr += 3;
                    ClonedPtr += 3;
                }
                Ptr += nOffset;
                ClonedPtr += nOffset;
            }
            ClonedImage.UnlockBits(ClonedBitmapData);
        }
        #endregion

        #region Pixels
        /// <summary>
        /// Gets the array of the values of each pixel
        /// </summary>
        /// <returns></returns>
        public unsafe double[,] PixelsValue()
        {
            IntPtr Scan0 = vBitmapData.Scan0;
            byte* p = (byte*)(void*)Scan0;
            int nOffset = vBitmapData.Stride - vBitmapData.Width * 3;
            double[,] Buffer = new double[3, 256];

            for (int y = 0; y < vBitmapData.Height; y++)
            {
                for (int x = 0; x < vBitmapData.Width; x++)
                {
                    Buffer[0, p[0]]++;
                    Buffer[1, p[1]]++;
                    Buffer[2, p[2]]++;
                    p += 3;
                }
                p += nOffset;
            }
            return Buffer;
        }
        #endregion
        #endregion

        #region Getter and Setter
        /// <summary>
        /// Get or set the image
        /// </summary>
        public Bitmap Bitmap
        {
            get { return vBitmap; }
            set
            {
                vPixelFormat = value.PixelFormat;
                vBitmap = (Bitmap)value.Clone();
            }
        }

        /// <summary>
        /// Get or set the pixel format of the image
        /// </summary>
        /// <remarks>Set the pixel format will force the opnening of the image in this format</remarks>
        public PixelFormat PixelFormat
        {
            get { return vPixelFormat; }
            set { vPixelFormat = value; }
        }
        #endregion
    }


}
